// Copyright 2018 Google LLC. All Rights Reserved.
/*
  Copyright (C) 2005-2018 Steven L. Scott

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
*/

#include "Models/HMM/Clickstream/Event.hpp"

namespace BOOM {
  namespace Clickstream {

    Event::Event(int event_type, const Ptr<CatKeyBase> &key)
        : MarkovData(uint(event_type), key) {}

    Event::Event(int event_type, const Ptr<Event> &prev)
        : MarkovData(uint(event_type), Ptr<MarkovData>(prev)) {}

    Event *Event::clone() const { return new Event(*this); }

  }  // namespace Clickstream
}  // namespace BOOM
